import * as Utils from "./Utils";
import { clamp } from "./Utils";
import * as NoiseGenerator from "./NoiseGenerator";
export class Glottis {
    constructor(sampleRate) {
        this.alwaysVoice = true;
        this.autoWobble = true;
        this.isTouched = false;
        this.targetTenseness = 0.6;
        this.targetFrequency = 140;
        this.vibratoAmount = 0.005;
        this.vibratoFrequency = 6;
        this.sampleCount = 0;
        this.intensity = 0;
        this.loudness = 1;
        this.smoothFrequency = 140;
        this.timeInWaveform = 0;
        this.newTenseness = 0.6;
        this.oldTenseness = 0.6;
        this.newFrequency = 140;
        this.oldFrequency = 140;
        this.sampleRate = sampleRate;
        this.aspirationNoiseSource = Utils.createFilteredNoiseSource(500, 0.5, sampleRate, 0x8000);
        this.setupWaveform(0);
    }
    step(lambda) {
        const time = this.sampleCount / this.sampleRate;
        if (this.timeInWaveform > this.waveformLength) {
            this.timeInWaveform -= this.waveformLength;
            this.setupWaveform(lambda);
        }
        const out1 = this.normalizedLFWaveform(this.timeInWaveform / this.waveformLength);
        const aspirationNoise = this.aspirationNoiseSource();
        const aspiration1 = this.intensity * (1 - Math.sqrt(this.targetTenseness)) * this.getNoiseModulator() * aspirationNoise;
        const aspiration2 = aspiration1 * (0.2 + 0.02 * NoiseGenerator.simplex1(time * 1.99));
        const out = out1 + aspiration2;
        this.sampleCount++;
        this.timeInWaveform += 1 / this.sampleRate;
        return out;
    }
    getNoiseModulator() {
        const voiced = 0.1 + 0.2 * Math.max(0, Math.sin(Math.PI * 2 * this.timeInWaveform / this.waveformLength));
        return this.targetTenseness * this.intensity * voiced + (1 - this.targetTenseness * this.intensity) * 0.3;
    }
    adjustParameters(deltaTime) {
        const delta = deltaTime * this.sampleRate / 512;
        const oldTime = this.sampleCount / this.sampleRate;
        const newTime = oldTime + deltaTime;
        this.adjustIntensity(delta);
        this.calculateNewFrequency(newTime, delta);
        this.calculateNewTenseness(newTime);
    }
    calculateNewFrequency(time, delta) {
        if (this.intensity == 0) {
            this.smoothFrequency = this.targetFrequency;
        }
        else if (this.targetFrequency > this.smoothFrequency) {
            this.smoothFrequency = Math.min(this.smoothFrequency * (1 + 0.1 * delta), this.targetFrequency);
        }
        else if (this.targetFrequency < this.smoothFrequency) {
            this.smoothFrequency = Math.max(this.smoothFrequency / (1 + 0.1 * delta), this.targetFrequency);
        }
        this.oldFrequency = this.newFrequency;
        this.newFrequency = Math.max(10, this.smoothFrequency * (1 + this.calculateVibrato(time)));
    }
    calculateNewTenseness(time) {
        this.oldTenseness = this.newTenseness;
        this.newTenseness = Math.max(0, this.targetTenseness + 0.1 * NoiseGenerator.simplex1(time * 0.46) + 0.05 * NoiseGenerator.simplex1(time * 0.36));
        if (!this.isTouched && this.alwaysVoice) {
            this.newTenseness += (3 - this.targetTenseness) * (1 - this.intensity);
        }
    }
    adjustIntensity(delta) {
        if (this.isTouched || this.alwaysVoice) {
            this.intensity += 0.13 * delta;
        }
        else {
            this.intensity -= 0.05 * delta;
        }
        this.intensity = clamp(this.intensity, 0, 1);
    }
    calculateVibrato(time) {
        let vibrato = 0;
        vibrato += this.vibratoAmount * Math.sin(2 * Math.PI * time * this.vibratoFrequency);
        vibrato += 0.02 * NoiseGenerator.simplex1(time * 4.07);
        vibrato += 0.04 * NoiseGenerator.simplex1(time * 2.15);
        if (this.autoWobble) {
            vibrato += 0.2 * NoiseGenerator.simplex1(time * 0.98);
            vibrato += 0.4 * NoiseGenerator.simplex1(time * 0.5);
        }
        return vibrato;
    }
    setupWaveform(lambda) {
        const frequency = this.oldFrequency * (1 - lambda) + this.newFrequency * lambda;
        const tenseness = this.oldTenseness * (1 - lambda) + this.newTenseness * lambda;
        this.waveformLength = 1 / frequency;
        this.loudness = Math.pow(Math.max(0, tenseness), 0.25);
        const rd = clamp(3 * (1 - tenseness), 0.5, 2.7);
        const ra = -0.01 + 0.048 * rd;
        const rk = 0.224 + 0.118 * rd;
        const rg = (rk / 4) * (0.5 + 1.2 * rk) / (0.11 * rd - ra * (0.5 + 1.2 * rk));
        const ta = ra;
        const tp = 1 / (2 * rg);
        const te = tp + tp * rk;
        const epsilon = 1 / ta;
        const shift = Math.exp(-epsilon * (1 - te));
        const delta = 1 - shift;
        const rhsIntegral = ((1 / epsilon) * (shift - 1) + (1 - te) * shift) / delta;
        const totalLowerIntegral = rhsIntegral - (te - tp) / 2;
        const totalUpperIntegral = -totalLowerIntegral;
        const omega = Math.PI / tp;
        const s = Math.sin(omega * te);
        const y = -Math.PI * s * totalUpperIntegral / (tp * 2);
        const z = Math.log(y);
        const alpha = z / (tp / 2 - te);
        const e0 = -1 / (s * Math.exp(alpha * te));
        this.alpha = alpha;
        this.e0 = e0;
        this.epsilon = epsilon;
        this.shift = shift;
        this.delta = delta;
        this.te = te;
        this.omega = omega;
    }
    normalizedLFWaveform(t) {
        let output;
        if (t > this.te) {
            output = (-Math.exp(-this.epsilon * (t - this.te)) + this.shift) / this.delta;
        }
        else {
            output = this.e0 * Math.exp(this.alpha * t) * Math.sin(this.omega * t);
        }
        return output * this.intensity * this.loudness;
    }
}
//# sourceMappingURL=Glottis.js.map