import { GlottisUi } from "./GlottisUi";
import { TractUi } from "./TractUi";
import * as GuiUtils from "./GuiUtils";
import { Button } from "./GuiUtils";
const projectUrl = "github.com/chdh/pink-tronbone-mod";
export class MainUi extends EventTarget {
    constructor(synthesizer, canvas) {
        super();
        this.mouseTouchCtr = 0;
        this.synthesizer = synthesizer;
        this.canvas = canvas;
        this.ctx = canvas.getContext("2d");
        this.glottis = synthesizer.glottis;
        this.glottisUi = new GlottisUi(synthesizer.glottis);
        this.tractUi = new TractUi(synthesizer.tract, synthesizer.tractShaper);
        this.touchesWithMouse = [];
        this.screen = 2;
        this.aboutButton = new Button(460, 392, 140, 30, "about...", true);
        this.alwaysVoiceButton = new Button(460, 428, 140, 30, "always voice", true);
        this.autoWobbleButton = new Button(460, 464, 140, 30, "pitch wobble", true);
        canvas.addEventListener("touchstart", (event) => this.touchStartEventHandler(event));
        canvas.addEventListener("touchmove", (event) => this.touchMoveEventHandler(event));
        canvas.addEventListener("touchend", (event) => this.touchEndEventHandler(event));
        canvas.addEventListener("touchcancel", (event) => this.touchEndEventHandler(event));
        canvas.addEventListener("mousedown", (event) => this.mouseDownEventHandler(event));
        document.addEventListener("mouseup", (event) => this.mouseUpEventHandler(event));
        document.addEventListener("mousemove", (event) => this.mouseMoveEventHandler(event));
        this.createBackgroundCanvas();
    }
    createBackgroundCanvas() {
        this.backCanvas = document.createElement("canvas");
        this.backCanvas.width = this.canvas.width;
        this.backCanvas.height = this.canvas.height;
        const ctx = this.backCanvas.getContext("2d");
        this.glottisUi.drawBackground(ctx);
        this.tractUi.drawBackground(ctx);
    }
    getTractTime() {
        return this.synthesizer.tract.time;
    }
    switchScreen(screen) {
        this.screen = screen;
        this.aboutButton.switchedOn = true;
        this.dispatchEvent(new CustomEvent("screen-switched"));
    }
    draw() {
        const ctx = this.ctx;
        ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
        ctx.drawImage(this.backCanvas, 0, 0);
        this.glottisUi.draw(ctx);
        this.tractUi.draw(ctx);
        this.alwaysVoiceButton.draw(ctx);
        this.autoWobbleButton.draw(ctx);
        this.aboutButton.draw(ctx);
        switch (this.screen) {
            case 2: {
                this.drawAboutScreen();
                break;
            }
            case 1: {
                this.drawInstructionsScreen();
                break;
            }
        }
    }
    drawAboutScreen() {
        const ctx = this.ctx;
        ctx.save();
        ctx.globalAlpha = 0.8;
        ctx.fillStyle = "white";
        ctx.rect(0, 0, ctx.canvas.width, ctx.canvas.height);
        ctx.fill();
        ctx.restore();
        this.drawAboutText();
    }
    drawAboutText() {
        const ctx = this.ctx;
        ctx.save();
        ctx.globalAlpha = 1;
        ctx.fillStyle = "#C070C6";
        ctx.strokeStyle = "#C070C6";
        ctx.font = "50px Arial";
        ctx.lineWidth = 3;
        ctx.textAlign = "center";
        const titleText = "P i n k   T r o m b o n e";
        ctx.strokeText(titleText, 300, 230);
        ctx.fillText(titleText, 300, 230);
        ctx.font = "28px Arial";
        ctx.fillText("bare-handed  speech synthesis", 300, 330);
        ctx.font = "20px Arial";
        ctx.restore();
    }
    drawInstructionsScreen() {
        const ctx = this.ctx;
        ctx.save();
        ctx.globalAlpha = 0.85;
        ctx.fillStyle = "white";
        ctx.rect(0, 0, ctx.canvas.width, ctx.canvas.height);
        ctx.fill();
        ctx.globalAlpha = 1;
        ctx.fillStyle = "#C070C6";
        ctx.strokeStyle = "#C070C6";
        ctx.font = "24px Arial";
        ctx.lineWidth = 2;
        ctx.textAlign = "center";
        ctx.font = "19px Arial";
        ctx.textAlign = "left";
        this.instructionsLine = 0;
        this.write("Sound is generated in the glottis (at the bottom left) then ");
        this.write("filtered by the shape of the vocal tract. The voicebox ");
        this.write("controls the pitch and intensity of the initial sound.");
        this.write("");
        this.write("Then, to talk:");
        this.write("");
        this.write("- move the body of the tongue to shape vowels");
        this.write("");
        this.write("- touch the oral cavity to narrow it, for fricative consonants");
        this.write("");
        this.write("- touch above the oral cavity to close it, for stop consonants");
        this.write("");
        this.write("- touch the nasal cavity to open the velum and let sound ");
        this.write("   flow through the nose.");
        this.write("");
        this.write("");
        this.write("(tap anywhere to continue)");
        this.instructionsLine = 18.8;
        ctx.textAlign = "left";
        this.write("This is the demo program for the pink-trombone-mod package.");
        ctx.save();
        ctx.fillStyle = "blue";
        ctx.globalAlpha = 0.6;
        this.write(projectUrl);
        ctx.restore();
        this.write("Pink Trombone was developed by Neil Thapen in 2017.");
        ctx.restore();
    }
    instructionsScreenHandleTouches(touches) {
        for (const touch of touches) {
            const p = GuiUtils.mapDomToCanvasCoordinates(this.canvas, touch.clientX, touch.clientY);
            this.instructionsScreenHandleTouch(p.x, p.y);
        }
    }
    instructionsScreenHandleTouch(x, y) {
        if ((x >= 35 && x <= 400) && (y >= 515 && y <= 540)) {
            window.location.href = "https://" + projectUrl;
        }
        else {
            this.switchScreen(0);
        }
    }
    write(text) {
        this.ctx.fillText(text, 50, 100 + this.instructionsLine * 22);
        this.instructionsLine += (text == "") ? 0.7 : 1;
    }
    buttonsHandleTouchStart(touch) {
        this.alwaysVoiceButton.handleTouchStart(touch);
        this.glottis.alwaysVoice = this.alwaysVoiceButton.switchedOn;
        this.autoWobbleButton.handleTouchStart(touch);
        this.glottis.autoWobble = this.autoWobbleButton.switchedOn;
        this.aboutButton.handleTouchStart(touch);
    }
    touchStartEventHandler(event) {
        event.preventDefault();
        switch (this.screen) {
            case 0: {
                this.processStartTouches(event.changedTouches);
                this.handleTouches();
                break;
            }
            case 2: {
                this.switchScreen(0);
                break;
            }
            case 1: {
                this.instructionsScreenHandleTouches(event.changedTouches);
                break;
            }
        }
    }
    processStartTouches(touches) {
        for (const touch of touches) {
            const appTouch = {};
            appTouch.startTime = this.getTractTime();
            appTouch.endTime = 0;
            appTouch.alive = true;
            appTouch.id = touch.identifier;
            this.updateAppTouchPosition(appTouch, touch.clientX, touch.clientY);
            this.touchesWithMouse.push(appTouch);
            this.buttonsHandleTouchStart(appTouch);
        }
    }
    updateAppTouchPosition(appTouch, clientX, clientY) {
        const p = GuiUtils.mapDomToCanvasCoordinates(this.canvas, clientX, clientY);
        appTouch.x = p.x;
        appTouch.y = p.y;
        appTouch.index = this.tractUi.getIndex(p.x, p.y);
        appTouch.diameter = this.tractUi.getDiameter(p.x, p.y);
    }
    getAppTouchById(id) {
        for (const appTouch of this.touchesWithMouse) {
            if (appTouch.id == id && appTouch.alive) {
                return appTouch;
            }
        }
        return undefined;
    }
    touchMoveEventHandler(event) {
        for (const touch of event.changedTouches) {
            const appTouch = this.getAppTouchById(touch.identifier);
            if (appTouch) {
                this.updateAppTouchPosition(appTouch, touch.clientX, touch.clientY);
            }
        }
        this.handleTouches();
    }
    touchEndEventHandler(event) {
        for (const touch of event.changedTouches) {
            const appTouch = this.getAppTouchById(touch.identifier);
            if (appTouch) {
                appTouch.alive = false;
                appTouch.endTime = this.getTractTime();
            }
        }
        this.handleTouches();
        if (this.screen == 0 && !this.aboutButton.switchedOn) {
            this.switchScreen(1);
        }
    }
    mouseDownEventHandler(event) {
        event.preventDefault();
        switch (this.screen) {
            case 0: {
                const appTouch = {};
                appTouch.startTime = this.getTractTime();
                appTouch.endTime = 0;
                appTouch.alive = true;
                appTouch.id = "mouse" + this.mouseTouchCtr++;
                this.updateAppTouchPosition(appTouch, event.clientX, event.clientY);
                this.mouseTouch = appTouch;
                this.touchesWithMouse.push(appTouch);
                this.buttonsHandleTouchStart(appTouch);
                this.handleTouches();
                break;
            }
            case 2: {
                this.switchScreen(0);
                break;
            }
            case 1: {
                const p = GuiUtils.mapDomToCanvasCoordinates(this.canvas, event.clientX, event.clientY);
                this.instructionsScreenHandleTouch(p.x, p.y);
                break;
            }
        }
    }
    mouseMoveEventHandler(event) {
        const appTouch = this.mouseTouch;
        if (!appTouch || !appTouch.alive) {
            return;
        }
        this.updateAppTouchPosition(appTouch, event.clientX, event.clientY);
        this.handleTouches();
    }
    mouseUpEventHandler(_event) {
        const appTouch = this.mouseTouch;
        if (!appTouch || !appTouch.alive) {
            return;
        }
        appTouch.alive = false;
        appTouch.endTime = this.getTractTime();
        this.handleTouches();
        if (this.screen == 0 && !this.aboutButton.switchedOn) {
            this.switchScreen(1);
        }
    }
    handleTouches() {
        this.removeOldTouches();
        if (this.screen == 0) {
            this.tractUi.handleTouches(this.touchesWithMouse, this.getTractTime());
            this.glottisUi.handleTouches(this.touchesWithMouse);
        }
    }
    removeOldTouches() {
        const time = this.getTractTime();
        for (let i = this.touchesWithMouse.length - 1; i >= 0; i--) {
            const appTouch = this.touchesWithMouse[i];
            if (!appTouch.alive && time > appTouch.endTime + 1) {
                this.touchesWithMouse.splice(i, 1);
            }
        }
    }
}
//# sourceMappingURL=MainUi.js.map